
var doc = window.document;

/*** Keyboard state ***/
var emod; /* ye event model */

var keypress_msg = "";  /* Debug messages from event handlers */
var keydown_msg = "";

/*** Keyboard constants ***/

/** These are from Firefox/Linux. Do they vary on other platforms? **/
var kc_up = 38;
var kc_down = 40;
var kc_left = 37;
var kc_right = 39;
var kc_delete = 46;
var kc_insert = 45;
var kc_home = 36;
var kc_end = 35;
var kc_pageup = 33;
var kc_pagedown = 34;
var kc_backspace = 8;
var kc_enter = 13;
var kc_escape = 27;
function kc_fn(x) { return 112+x-1; } /* Function keys */


/** Debug IO **/
function mywrite(x) {
  var body = doc.getElementById("debugstream");
  var paragraph = doc.createElement("P");
  var text = doc.createTextNode(x);
  paragraph.appendChild(text);
  body.appendChild(paragraph);
  
  //body.innerHTML += x + "<br>";
}

function refresh_keybuf () {
  doc.getElementById("keybuf").innerHTML = keypress_msg + "<BR>" + keydown_msg;
}

function assertion_error (message) { mywrite("Assertion failure: " + message); throw(message); return null;}
function assertfn (fn) { if (!fn()) assertion_error(fn.toString()); }
function assertnn (x, descr) { return (x == null)? assertion_error(descr + " is null") : x; }

/**** Editor ****/

/** Editor protocol:
    The editor state consists of a doubly-linked list of spans within the DOM,
    with which methods implementing editing operations are (directly or indirectly)
    associated.

    Rather than accessing the links or properties directly, use the following
    functions:
    
    edn_next (node)
    edn_prev (node)
    edn_insertahead (node,newnode)
    edn_insertbehind (node,newnode)
    edn_unlink (node)
    edn_keyhandler (node)
    edn_text (node)
    edn_settext (node, newtext)

    Internals:
    _edn_setnext (node)
    _edn_setprev (node)
**/

var curcontainer;
var rootnode; // These should really be wrapped in an "editor instance", but what-fucking-ever.
var curnode;

function edn_next (node) { return assertnn(node, "node (for next)").calc_nextnode; }
function edn_prev (node) { return assertnn(node, "node (for prev)").calc_prevnode; }
function _edn_setnext (node, val) { node.calc_nextnode = val; }
function _edn_setprev (node, val) { node.calc_prevnode = val; }
function edn_keyhandler (node) { return node.calc_keypress; }

function edn_assert_invariants (node) {
  assertfn(function(){return node != null;});
  if (edn_next(node) != null) assertfn(function(){return (node == edn_prev(edn_next(node)));});  
  if (edn_prev(node) != null) assertfn(function(){return (node == edn_next(edn_prev(node)));});
  else assertfn(function(){return (rootnode == node);});
}

function edn_text (node) { var t=node.edn_text; assertfn(function(){ return !t || (t.length==1)}); return t; }
function edn_settext (node, newtext) { assertnn(node).edn_text = newtext; }

function edn_insertahead (node, newnode) {
  edn_assert_invariants(node);
  if (edn_next(node)) {
    _edn_setnext(newnode, edn_next(node));
    _edn_setprev(edn_next(node), newnode);
  }
  _edn_setnext(node, newnode);
  _edn_setprev(newnode, node);
  edn_assert_invariants(node);
  edn_assert_invariants(newnode);
  assertion_error("edn_insertahead: Fixme!");
}

function edn_insertbehind (node, newnode) {
  edn_assert_invariants(node);
  var oldprev = edn_prev(node);
  if (oldprev) {
    _edn_setprev(newnode, oldprev);
    _edn_setnext(oldprev, newnode);
  } else rootnode = newnode;
  _edn_setnext(newnode, node);
  _edn_setprev(node, newnode);
  edn_assert_invariants(node);
  edn_assert_invariants(newnode);
  if (oldprev) edn_assert_invariants(oldprev);
  node.parentNode.insertBefore(newnode, node);
}

function edn_unlink (node) {
  var prev = edn_prev(node);
  var next = edn_next(node);

  if (!prev && !next) assertion_error("Can't unlink last editor node!");
  
  edn_assert_invariants(node);
  if (prev) _edn_setnext(prev, next);
  else rootnode = next;
  if (next) _edn_setprev(next, prev);
  node.parentNode.removeChild(node);
  _edn_setnext(node, null);
  _edn_setprev(node, null);
  if (prev) edn_assert_invariants(prev);
  if (next) edn_assert_invariants(next);
}

function anoint_editor_node (node) {
  _edn_setnext(node, null);
  _edn_setprev(node, null);
  node.calc_keypress = charcell_keypress;
  //node.style.minWidth = "4px";  // Why can't I make this work?
}

function charcell_insert (node, charstring) {
  var newnode = doc.createElement("span");
  var text = doc.createTextNode(charstring);
  edn_settext(newnode, charstring);
  anoint_editor_node(newnode);
  newnode.appendChild(text);
  newnode.calc_keypress = charcell_keypress;
  edn_insertbehind(curnode, newnode);
  return node;
}

function charcell_keypress (node, control, meta, kc, cc) {
  if (!control && !meta && (cc!=0)) {
    return charcell_insert(node, String.fromCharCode(cc));
  } else {
    if (!control && !meta) {
      switch (kc) {
      case kc_left:  return edn_prev(curnode) || rootnode;
      case kc_right: return edn_next(curnode) || curnode;
      case kc_backspace: 
        if (edn_prev(curnode)) edn_unlink(edn_prev(curnode));
        return curnode;
      case kc_delete:
        var next = edn_next(curnode);
        if (next) edn_unlink(curnode);
        return next || curnode;
      case kc_enter: do_eval_buffer(); return curnode;
      }
    }
    return node;
  }
}

function node_setstyle_default (node) {
  node.style.backgroundColor = null;
  node.style.color = "black";
}

function node_setstyle_editing (node) {
  node.style.backgroundColor = "black";
  node.style.color = "white";
}


/* Node editing protocol:
   Every node should should have a 'calc_keypress' function.
   ???
*/

/** This is called for every keystroke by the event handlers: **/
function keypress (control, meta, kc, cc) {
  var nextnode = edn_keyhandler(curnode)(curnode, control, meta, kc, cc);
  if (nextnode != curnode) {
    node_setstyle_default(curnode);
    node_setstyle_editing(nextnode);
    curnode = nextnode;
  }
}

function test_keypress (control, meta, kc, cc) {
  var hotkey = control || meta;
  var stream = doc.getElementById("output");
  if ((cc > 0) && !hotkey) { /* Regular character input */
    stream.innerHTML += String.fromCharCode(cc);
  } else {        /* Special character or modifier chord */

    if (!hotkey) {
      if (kc == kc_enter) stream.innerHTML += "<BR>";
    }    
  }
}

function do_eval_buffer () {
  node_setstyle_default(curnode);

  parse_and_eval(rootnode);

  // Create paragraph for the new interaction area:
  var stream = doc.getElementById("output");
  var container = doc.createElement("P");
  stream.appendChild(container);
  
  // Create new root node:
  var newroot = doc.createElement("span");
  //var filltext = doc.createTextNode("&nbsp;");
  //newroot.appendChild(filltext);
  newroot.innerHTML="&nbsp;";
  container.appendChild(newroot);
  anoint_editor_node(newroot);
  node_setstyle_editing(newroot);
  
  // Swap in the new editor state:
  curcontainer = container;
  rootnode = curnode = newroot;    
}


// I ripped and adapted this from a webmonkey tutorial.
// I can almost write the whole line editor in the space this piece of shit
// function spends jerking itself off with pointless variables.
function makeKeyboardHandler (dbgprinter) {
  return function (e) {
   var output_string;

   switch (emod) {
   case "IE4+":
     var alt = false;
     var ctrl = false;
     var shift = false;

      e = window.event;
      /*C*/
      alt = (e.altKey) ? true : false;
      ctrl = (e.ctrlKey) ? true : false;
      shift = (e.shiftKey) ? true : false;

      output_string = " alt: "+alt+"\n ctrl: "+ctrl+"\n shift: "+shift;

      /*D*/
      if ((e.keyCode<16)||(e.keyCode>18)) {
         output_string += "\n 'keyCode' attribute: "+e.keyCode;
	 mywrite(output_string);
         keypress(ctrl, String.fromCharCode(e.keyCode));
         alt = false;
         ctrl = false;
         shift = false; }
   break;
   case "W3C":
     var alt = false;
     var ctrl = false;
     var shift = false;

      alt = (e.altKey) ? true : false;
      ctrl = (e.ctrlKey) ? true : false;
      shift = (e.shiftKey) ? true : false;

      // Okay, this is no longer "W3C". It is now "Firefox". <-- FIXME
      // Why? 'which' is useless. 

      output_string = (alt?"[alt] ":"") + (ctrl?"[control] ":"") + (shift?"[shift] ":"");

      if ((e.which<16)||(e.which>18))
      {
        //output_string += " e.which="+e.which;
        output_string = "e.keyCode=" + e.keyCode + " e.charCode=" + e.charCode + output_string;
	 //mywrite(output_string);
         dbgprinter(output_string);
         keypress(ctrl, alt || e.metaKey, e.keyCode, e.charCode);
      }
   break; }

   /*G*/
   return false;
}}

function handle_onload(e)
{
   /*get the event model*/
   emod = (e) ? (e.eventPhase) ? "W3C" : "NN4" : (window.event) ? "IE4+" : "unknown";
   mywrite("Event model: " + emod);

   rootnode = curnode = doc.getElementById("editroot"); // TEMP
   curcontainer = doc.getElementById("firsteditor");
   anoint_editor_node(rootnode);
   node_setstyle_editing(rootnode);

   //document.onkeydown = onKeyDownH;
   //document.onkeydown = makeKeyboardHandler(function(s){keydown_msg="keydown: " + s; refresh_keybuf();});
   document.onkeypress = makeKeyboardHandler(/* debug printer: */function(s){keypress_msg="keypress: " + s; refresh_keybuf();});
   return true;
}

/*define the event handler for the onload event*/
window.onload = handle_onload;
